@@ -3,12 +3,11 @@ use warnings;
 
 package XML::Rabbit::Trait::XPath;
 {
-  $XML::Rabbit::Trait::XPath::VERSION = '0.2.1';
+  $XML::Rabbit::Trait::XPath::VERSION = '0.3.0';
 }
 use Moose::Role;
 use Moose::Util::TypeConstraints;
 use Perl6::Junction ();
-use Data::Visitor::Callback ();
 
 # ABSTRACT: Base role for other xpath traits
 
@@ -114,12 +113,36 @@ sub _resolve_class {
     # Figure out classes mentioned in type constraint (isa)
     my %classes;
     if ( $self->has_type_constraint ) {
-        Data::Visitor::Callback->new({
-            object => 'visit_ref',
-            'Moose::Meta::TypeConstraint::Union'         => sub { return $_[1]->type_constraints; },
-            'Moose::Meta::TypeConstraint::Class'         => sub { $classes{ $_[1]->class } = 1; return $_[1]; },
-            'Moose::Meta::TypeConstraint::Parameterized' => sub { return $_[1]->type_parameter; },
-        })->visit($self->type_constraint);
+
+        # Visit each part of the TC in turn, deconstruct and extract class names
+        my @to_visit = $self->type_constraint;
+        while (my $item = shift @to_visit) {
+            next unless blessed $item;
+            if ( $item->isa('Moose::Meta::TypeConstraint::Class') ) {
+                # Just a plain string that is a class
+                $classes{ $item->class } = 1;
+            }
+            elsif ( $item->isa('Moose::Meta::TypeConstraint::Union') ) {
+                # ArrayRef[Something|SomethingElse]
+                push @to_visit, @{ $item->type_constraints };
+            }
+            elsif ( $item->isa('Moose::Meta::TypeConstraint::Parameterized') ) {
+                # Maybe[Something], ArrayRef[Something] or HashRef[Something]
+                push @to_visit, $item->type_parameter;
+            }
+            elsif ( $item->isa('Moose::Meta::TypeConstraint::Parameterizable') ) {
+                # Built-in like ArrayRef or HashRef without a parameter
+                next;
+            }
+            else {
+                #warn("Unsupported TC detected: $item");
+            }
+        }
+
+        # Some debugging aid
+        #my $tc_name = $self->type_constraint->name;
+        #my $found = join(", ", sort keys %classes);
+        #warn("Classes found in TC '$tc_name': $found\n");
     }
 
     # RT#81519: The above code was supposed to not return duplicate class
@@ -130,9 +153,7 @@ sub _resolve_class {
     my @classes = keys %classes;
 
     # Runtime load each class
-    foreach my $class ( @classes ) {
-        Class::MOP::load_class($class);
-    }
+    Class::MOP::load_class($_) for @classes;
 
     # Return 0 if multiple classes found,
     # _create_instance() must use $self->isa_map to resolve class name
@@ -231,7 +252,7 @@ XML::Rabbit::Trait::XPath - Base role for other xpath traits
 
 =head1 VERSION
 
-version 0.2.1
+version 0.3.0
 
 =head1 SYNOPSIS
 
@@ -299,7 +320,7 @@ Robin Smidsrød <robin@smidsrod.no>
 
 =head1 COPYRIGHT AND LICENSE
 
-This software is copyright (c) 2012 by Robin Smidsrød.
+This software is copyright (c) 2013 by Robin Smidsrød.
 
 This is free software; you can redistribute it and/or modify it under
 the same terms as the Perl 5 programming language system itself.
